/*
 * Dinning philosophers problem with the parity strategy.
 *
 * Run as:
 *  (1) spin -a -DN=N philosophers.pml
 *  (2) cc -O2 pan.c -o pan
 *  (3) ./pan
 *
 * Time has been measured with Bash's time and memory usage with glibc's memusage.
 */
 

// Number of philosophers
#ifndef N
	#define N 5
#endif

// Boolean availability vector of forks
byte fork[N] = 1;
// Number of forks obtained by each philosopher
byte nforks[N] = 0;

active [N] proctype philosopher() {

	// Index of the left fork (_pid is the philosopher number
	// and also the index of the right fork, set by SPIN from 0 to N-1)
	byte left = (_pid > 0 -> _pid - 1 : N - 1);

	do
	:: atomic {
		if
		// Odd, take the right fork first
		:: nforks[_pid] == 0 && _pid % 2 == 1 && fork[_pid] == 1 -> fork[_pid] = 0; nforks[_pid] = 1;
		// Even, take the left fork first
		:: nforks[_pid] == 0 && _pid % 2 == 0 && fork[left] == 1 -> fork[left] = 0; nforks[_pid] = 1;
		// Take the other fork
		:: nforks[_pid] == 1 && _pid % 2 == 1 && fork[left] == 1 -> fork[left] = 0; nforks[_pid] = 2;
		:: nforks[_pid] == 1 && _pid % 2 == 0 && fork[_pid] == 1 -> fork[_pid] = 0; nforks[_pid] = 2;
		// Release both forks
		:: nforks[_pid] == 2 -> fork[_pid] = 0; fork[left] = 0; nforks[_pid] = 0;
		fi
	}
	od

}

// LTL properties (for different number of philosophers)

#if (N == 3)
ltl no_starv { <> (nforks[0] == 2 || nforks[1] == 2 || nforks[2] == 2) }
#endif

#if (N == 5)
ltl no_starv { <> (nforks[0] == 2 || nforks[1] == 2 || nforks[2] == 2 || nforks[3] == 2 || nforks[4] == 2) }
#endif

#if (N == 7)
ltl no_starv { <> (nforks[0] == 2 || nforks[1] == 2 || nforks[2] == 2 || nforks[3] == 2 || nforks[4] == 2 || nforks[5] == 2 || nforks[6] == 2) }
#endif

#if (N == 11)
ltl no_starv { <> (nforks[0] == 2 || nforks[1] == 2 || nforks[2] == 2 || nforks[3] == 2 || nforks[4] == 2 || nforks[5] == 2 || nforks[6] == 2 || nforks[7] == 2 || nforks[8] == 2 || nforks[9] == 2 || nforks[10] == 2) }
#endif

#if (N == 13)
ltl no_starv { <> (nforks[0] == 2 || nforks[1] == 2 || nforks[2] == 2 || nforks[3] == 2 || nforks[4] == 2 || nforks[5] == 2 || nforks[6] == 2 || nforks[7] == 2 || nforks[8] == 2 || nforks[9] == 2 || nforks[10] == 2 || nforks[11] == 2 || nforks[12] == 2) }
#endif

#if (N == 17)
ltl no_starv { <> (nforks[0] == 2 || nforks[1] == 2 || nforks[2] == 2 || nforks[3] == 2 || nforks[4] == 2 || nforks[5] == 2 || nforks[6] == 2 || nforks[7] == 2 || nforks[8] == 2 || nforks[9] == 2 || nforks[10] == 2 || nforks[11] == 2 || nforks[12] == 2 || nforks[13] == 2 || nforks[14] == 2 || nforks[15] == 2 || nforks[16] == 2) }
#endif

#if (N == 23)
ltl no_starv { <> (nforks[0] == 2 || nforks[1] == 2 || nforks[2] == 2 || nforks[3] == 2 || nforks[4] == 2 || nforks[5] == 2 || nforks[6] == 2 || nforks[7] == 2 || nforks[8] == 2 || nforks[9] == 2 || nforks[10] == 2 || nforks[11] == 2 || nforks[12] == 2 || nforks[13] == 2 || nforks[14] == 2 || nforks[15] == 2 || nforks[16] == 2 || nforks[17] == 2 || nforks[18] == 2 || nforks[19] == 2 || nforks[20] == 2 || nforks[21] == 2 || nforks[22] == 2) }
#endif

#if (N == 25)
ltl no_starv { <> (nforks[0] == 2 || nforks[1] == 2 || nforks[2] == 2 || nforks[3] == 2 || nforks[4] == 2 || nforks[5] == 2 || nforks[6] == 2 || nforks[7] == 2 || nforks[8] == 2 || nforks[9] == 2 || nforks[10] == 2 || nforks[11] == 2 || nforks[12] == 2 || nforks[13] == 2 || nforks[14] == 2 || nforks[15] == 2 || nforks[16] == 2 || nforks[17] == 2 || nforks[18] == 2 || nforks[19] == 2 || nforks[20] == 2 || nforks[21] == 2 || nforks[22] == 2 || nforks[23] == 2 || nforks[24] == 2) }
#endif

#if (N == 27)
ltl no_starv { <> (nforks[0] == 2 || nforks[1] == 2 || nforks[2] == 2 || nforks[3] == 2 || nforks[4] == 2 || nforks[5] == 2 || nforks[6] == 2 || nforks[7] == 2 || nforks[8] == 2 || nforks[9] == 2 || nforks[10] == 2 || nforks[11] == 2 || nforks[12] == 2 || nforks[13] == 2 || nforks[14] == 2 || nforks[15] == 2 || nforks[16] == 2 || nforks[17] == 2 || nforks[18] == 2 || nforks[19] == 2 || nforks[20] == 2 || nforks[21] == 2 || nforks[22] == 2 || nforks[23] == 2 || nforks[24] == 2 || nforks[25] == 2 || nforks[26] == 2) }
#endif

#if (N == 29)
ltl no_starv { <> (nforks[0] == 2 || nforks[1] == 2 || nforks[2] == 2 || nforks[3] == 2 || nforks[4] == 2 || nforks[5] == 2 || nforks[6] == 2 || nforks[7] == 2 || nforks[8] == 2 || nforks[9] == 2 || nforks[10] == 2 || nforks[11] == 2 || nforks[12] == 2 || nforks[13] == 2 || nforks[14] == 2 || nforks[15] == 2 || nforks[16] == 2 || nforks[17] == 2 || nforks[18] == 2 || nforks[19] == 2 || nforks[20] == 2 || nforks[21] == 2 || nforks[22] == 2 || nforks[23] == 2 || nforks[24] == 2 || nforks[25] == 2 || nforks[26] == 2 || nforks[27] == 2 || nforks[28] == 2) }
#endif
